from enum import Enum


class ConnectionStatus(Enum):
    """Connection status of the NibeGW connection."""

    UNKNOWN = "unknown"
    INITIALIZING = "initializing"
    LISTENING = "listening"
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"

    def __str__(self):
        return self.value


class ConnectionStatusMixin:
    CONNECTION_STATUS_EVENT = "connection_status"
    _status: ConnectionStatus = ConnectionStatus.UNKNOWN

    @property
    def status(self) -> ConnectionStatus:
        """Get the current connection status"""
        return self._status

    @status.setter
    def status(self, status: ConnectionStatus):
        if status != self._status:
            self._status = status
            self.notify_event_listeners(self.CONNECTION_STATUS_EVENT, status=status)

    def notify_event_listeners(self, *args, **kwargs):
        pass



import asyncio
from collections.abc import Mapping, MutableMapping
from importlib.resources import files, open_text
import json
import logging
import re

import pandas
from slugify import slugify

from nibe.heatpump import HeatPump, Model

logger = logging.getLogger("nibe")


def update_dict(d: MutableMapping, u: Mapping, removeExplicitNulls: bool) -> Mapping:
    for k, v in u.items():
        if v is None and removeExplicitNulls:
            d.pop(k, None)
        elif isinstance(v, Mapping):
            update_dict(d.setdefault(k, {}), v, removeExplicitNulls)
        else:
            d[k] = v

    return d


class CSVConverter:
    data: pandas.DataFrame

    def __init__(self, in_file, out_file, extensions):
        self.in_file = in_file
        self.out_file = out_file
        self.extensions = extensions

    def convert(self):
        self._read_csv()

        self._unifi_column_names()

        self._update_index()

        self._fix_data_soft_hyphens()

        self._make_name_using_slugify()

        self._fix_data_unit_column()

        self._fix_data_types()

        self._fix_data_size_column()

        self._replace_mode_with_boolean_write_parameter()

        self._unset_equal_min_max_default_values()

        self._make_mapping_parameter()

        self._export_to_file()

    def _make_dict(self):
        return {index: row.dropna().to_dict() for index, row in self.data.iterrows()}

    def _make_mapping_parameter(self):
        if "info" not in self.data:
            return

        re_mapping = re.compile(
            r"(?P<value>\d+|I)\s*=\s*(?P<key>(?:[\w +.-]+[\w]\b[+]?(?! *=)))",
            re.IGNORECASE,
        )
        mappings = (
            self.data["info"]
            .where(~self.data["info"].str.contains("encoded"))
            .str.extractall(re_mapping)
        )
        mappings["value"] = mappings["value"].str.replace("I", "1").astype("str")
        mappings = mappings.reset_index("match", drop=True)
        self.data["mappings"] = pandas.Series(
            {
                str(k): self._make_mapping_series(g)
                for k, g in mappings.groupby("value", level=0)
            }
        ).where(self._is_mapping_allowed)

    def _is_mapping_allowed(self, s):
        return self.data["factor"] == 1

    def _make_mapping_series(self, g):
        return g.set_index("value", drop=True)["key"].drop_duplicates()

    def _unset_equal_min_max_default_values(self):
        valid_min_max = self.data["min"] != self.data["max"]
        self.data["min"] = self.data["min"].where(valid_min_max)
        self.data["max"] = self.data["max"].where(valid_min_max)
        self.data["default"] = self.data["default"].where(valid_min_max)

    def _fix_data_types(self):
        self.data["unit"] = self.data["unit"].astype("string")
        self.data["title"] = self.data["title"].astype("string")
        if "info" in self.data:
            self.data["info"] = self.data["info"].astype("string")
        self.data["size"] = self.data["size"].astype("string")
        self.data["name"] = self.data["name"].astype("string")
        self.data["factor"] = self.data["factor"].astype("int")
        self.data["min"] = self.data["min"].astype("float")
        self.data["max"] = self.data["max"].astype("float")
        self.data["default"] = self.data["default"].astype("float")

    def _fix_data_size_column(self):
        mapping = {
            "1.0": "s8",
            "2.0": "s16",
            "3.0": "s32",
            "4.0": "u8",
            "5.0": "u16",
            "6.0": "u32",
            "s8": "s8",
            "s16": "s16",
            "s32": "s32",
            "u8": "u8",
            "u16": "u16",
            "u32": "u32",
        }

        size = self.data["size"].map(mapping)

        invalid_size = size.isna()
        if any(invalid_size):
            logger.warning(
                "Invalid size data replaced with u16:\n%s", self.data[invalid_size]
            )
            size[invalid_size] = "u16"

        self.data["size"] = size

    def _fix_data_unit_column(self):
        self.data["unit"] = (
            self.data["unit"].replace(r"^\s*$", pandas.NA, regex=True).str.strip()
        )

    def _fix_data_soft_hyphens(self):
        self.data["title"] = self.data["title"].str.replace("\xad", "")

    def _make_name_using_slugify(self):
        ids = pandas.Series(self.data.index, index=self.data.index)
        self.data["name"] = self.data["title"].combine(
            ids, lambda title, id_: slugify(f"{title}-{id_}")
        )

    def _replace_mode_with_boolean_write_parameter(self):
        if "mode" in self.data:
            self.data["mode"] = self.data["mode"].str.strip().astype("string")

            self.data["write"] = self.data["mode"].map(
                lambda x: True if x == "R/W" else pandas.NA
            )
            del self.data["mode"]

    def _unifi_column_names(self):
        self.data.columns = map(str.lower, self.data.columns)
        self.data.rename(
            columns={
                "division factor": "factor",
                "size of variable": "size",
                "min value": "min",
                "max value": "max",
                "default value": "default",
            },
            inplace=True,
        )

    def _read_csv(self):
        with open(self.in_file, encoding="latin1") as f:
            modbus_manager = f.readline().startswith("ModbusManager")

        if modbus_manager:
            self.data = pandas.read_csv(
                self.in_file,
                sep=";",
                skiprows=4,
                encoding="latin1",
                index_col=False,
                skipinitialspace=True,
            )
        else:
            self.data = pandas.read_csv(
                self.in_file,
                sep="\t",
                skiprows=0,
                encoding="utf-8",
                index_col=False,
                skipinitialspace=True,
                na_values="-",
            )

    def _update_index(self):
        def calculate_number(row):
            register_type: str = row["register type"]
            location = 1 + int(row["register"])
            if register_type == "MODBUS_COIL":
                return str(int(location))
            if register_type == "MODBUS_DISCRETE_INPUT":
                return str(10000 + int(location))
            if register_type == "MODBUS_INPUT_REGISTER":
                return str(30000 + int(location))
            if register_type == "MODBUS_HOLDING_REGISTER":
                return str(40000 + int(location))
            return None

        if "register type" in self.data:
            id_prefixed = self.data.loc[
                self.data["title"].str.startswith("id:", na=True)
            ]
            if len(id_prefixed) > 0:
                logger.warning(
                    "Ignoring unnamed and often duplicated rows:\n%s",
                    id_prefixed,
                )
                self.data.drop(id_prefixed.index, inplace=True)

            self.data["id"] = self.data.apply(calculate_number, axis=1)

            self.data["mode"] = self.data["register type"].map(
                lambda x: "R/W"
                if x in ("MODBUS_HOLDING_REGISTER", "MODBUS_COIL")
                else "R"
            )

            del self.data["register type"]
            del self.data["register"]

        self.data["id"] = self.data["id"].astype("string")

        self.data = self.data.set_index("id")

    def _convert_series_to_dict(self, o):
        if isinstance(o, pandas.Series):
            return o.sort_index(key=lambda i: i.astype(int)).to_dict()

        raise TypeError(f"Object of type {type(o)} is not JSON serializable")

    def _export_to_file(self):
        o = self._make_dict()
        update_dict(o, self.extensions, True)
        with open(self.out_file, "w", encoding="utf-8") as fh:
            json.dump(o, fh, indent=2, default=self._convert_series_to_dict)
            fh.write("\n")


async def run():
    with open_text("nibe.data", "extensions.json") as fp:
        all_extensions = json.load(fp)

    for in_file in files("nibe.data").glob("*.csv"):
        out_file = in_file.with_suffix(".json")

        extensions = {}
        for extra in all_extensions:
            if out_file.name not in extra["files"]:
                continue
            update_dict(extensions, extra["data"], False)

        logger.info(f"Converting {in_file} to {out_file}")
        try:
            CSVConverter(in_file, out_file, extensions).convert()

            await _validate(out_file)

            logger.info(f"Converted {in_file} to {out_file}")
        except Exception as ex:
            logger.exception("Failed to convert %s: %s", in_file, ex)


async def _validate(out_file):
    model = Model.CUSTOM
    model.data_file = out_file
    hp = HeatPump(model)
    await hp.initialize()


def main():
    logging.basicConfig(level=logging.INFO)
    asyncio.run(run())


if __name__ == "__main__":
    main()



from contextlib import nullcontext
import datetime

import pytest

from nibe.coil import Coil, CoilData
from nibe.connection.nibegw import CoilDataEncoderNibeGw
from nibe.exceptions import DecodeException, EncodeException, ValidationError
from nibe.parsers import swapwords


def test_swapwords():
    assert swapwords(b"abcd") == b"cdab"
    assert swapwords(b"ab") == b"ab"


def test_create():
    coil = Coil(123, "test_name", "test_title", "u8", unknown="some other")

    assert coil.address == 123
    assert coil.name == "test_name"
    assert coil.title == "test_title"
    assert coil.other["unknown"] == "some other"


@pytest.fixture
def encoder():
    return CoilDataEncoderNibeGw()


@pytest.fixture
def encoder_word_swap_false():
    return CoilDataEncoderNibeGw(word_swap=False)


@pytest.fixture
def encoder_word_swap():
    return CoilDataEncoderNibeGw(word_swap=True)


# Signed 8-bit
@pytest.fixture
def coil_signed_s8():
    return Coil(48739, "cool-offset-s1-48739", "Cool offset S1", "s8")


@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\xfc\x00\x00\x00", -4),
        (b"\xfc\x00", -4),
        (b"\xfc", -4),
        (b"\xff", -1),
        (b"\x7f", 127),
        (b"\x81", -127),
        (b"\x80", None),
    ],
)
def test_signed_s8_decode(
    raw_value,
    value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_signed_s8: Coil,
):
    assert encoder_word_swap_false.decode(coil_signed_s8, raw_value) == CoilData(
        coil_signed_s8, value
    )


def test_signed_s8_encode(
    encoder_word_swap_false: CoilDataEncoderNibeGw, coil_signed_s8: Coil
):
    coil_data = CoilData(coil_signed_s8, -4)
    assert encoder_word_swap_false.encode(coil_data) == b"\xfc\x00\x00\x00"


@pytest.mark.parametrize("value", [(128), (None), (-129)])
def test_signed_s8_encode_exceptions(
    value, encoder_word_swap_false: CoilDataEncoderNibeGw, coil_signed_s8: Coil
):
    with pytest.raises(EncodeException):
        coil_data = CoilData(coil_signed_s8, value)
        encoder_word_swap_false.encode(coil_data)


# Unsigned 8-bit
@pytest.fixture
def coil_unsigned_u8():
    return Coil(123, "test", "test", "u8")


def test_validate_error(coil_unsigned_u8: Coil):
    coil_data = CoilData(coil_unsigned_u8, "fail")
    with pytest.raises(ValidationError):
        coil_data.validate()


@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x01\x00\x00\x00", 1),
        (b"\x01\x00", 1),
        (b"\x01", 1),
        (b"\xff", None),
        (b"\xff\xff", None),
    ],
)
def test_unsigned_s8_decode(
    raw_value,
    value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_unsigned_u8: Coil,
):
    assert encoder_word_swap_false.decode(coil_unsigned_u8, raw_value) == CoilData(
        coil_unsigned_u8, value
    )


@pytest.mark.parametrize(
    "value, raw_value",
    [
        (1, b"\x01\x00\x00\x00"),
        (255, b"\xff\x00\x00\x00"),
    ],
)
def test_unsigned_s8_encode(
    value,
    raw_value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_unsigned_u8: Coil,
):
    coil_data = CoilData(coil_unsigned_u8, value)
    assert encoder_word_swap_false.encode(coil_data) == raw_value


def test_unsigned_s8_encode_exception(
    encoder_word_swap_false: CoilDataEncoderNibeGw, coil_unsigned_u8: Coil
):
    with pytest.raises(EncodeException):
        coil_data = CoilData(coil_unsigned_u8, 256)
        encoder_word_swap_false.encode(coil_data)


# Unsigned 8-bit with word swap
@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x01\x00\x00\x00", 1),
        (b"\x01\x00", 1),
        (b"\x01", 1),
        (b"\xff", None),
        (b"\xff\xff", None),
    ],
)
def test_unsigned_s8_decode_word_swap(
    raw_value, value, encoder_word_swap: CoilDataEncoderNibeGw, coil_unsigned_u8: Coil
):
    assert encoder_word_swap.decode(coil_unsigned_u8, raw_value) == CoilData(
        coil_unsigned_u8, value
    )


@pytest.mark.parametrize(
    "value, raw_value",
    [
        (1, b"\x01\x00\x00\x00"),
        (255, b"\xff\x00\x00\x00"),
    ],
)
def test_unsigned_s8_encode_word_swap(
    value, raw_value, encoder_word_swap: CoilDataEncoderNibeGw, coil_unsigned_u8: Coil
):
    coil_data = CoilData(coil_unsigned_u8, value)
    assert encoder_word_swap.encode(coil_data) == raw_value


def test_unsigned_s8_encode_exception_word_swap(
    encoder_word_swap: CoilDataEncoderNibeGw, coil_unsigned_u8: Coil
):
    with pytest.raises(EncodeException):
        coil_data = CoilData(coil_unsigned_u8, 256)
        encoder_word_swap.encode(coil_data)


# Signed 16-bit
@pytest.fixture
def coil_signed_s16():
    return Coil(123, "test", "test", "s16", factor=10, min=50, max=300)


def test_signed_s16_attributes(coil_signed_s16: Coil):
    assert coil_signed_s16.min == 5.0
    assert coil_signed_s16.max == 30.0

    assert coil_signed_s16.raw_min == 50
    assert coil_signed_s16.raw_max == 300

    assert not coil_signed_s16.is_boolean
    assert not coil_signed_s16.is_writable


@pytest.mark.parametrize(
    "value, expected_raises",
    [
        (5.0, nullcontext()),
        (30.0, nullcontext()),
        (4.9, pytest.raises(ValidationError)),
        (30.1, pytest.raises(ValidationError)),
    ],
)
def test_signed_s16_is_valid(value, expected_raises, coil_signed_s16: Coil):
    coil_data = CoilData(coil_signed_s16, value)
    with expected_raises:
        coil_data.validate()


@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x97\x00", 15.1),
        (b"\x00\x80", None),
    ],
)
def test_signed_s16_decode(
    raw_value,
    value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_signed_s16: Coil,
):
    actual_coil_data = encoder_word_swap_false.decode(coil_signed_s16, raw_value)
    assert actual_coil_data == CoilData(coil_signed_s16, value)
    assert isinstance(actual_coil_data.value, float) or actual_coil_data.value is None


@pytest.mark.parametrize(
    "raw_value",
    [
        (b"\x31\x00"),
        (b"\x2d\x10"),
    ],
)
def test_signed_s16_decode_exception(
    raw_value, encoder_word_swap_false: CoilDataEncoderNibeGw, coil_signed_s16: Coil
):
    with pytest.raises(DecodeException):
        encoder_word_swap_false.decode(coil_signed_s16, raw_value)


def test_signed_s16_encode(
    encoder_word_swap_false: CoilDataEncoderNibeGw, coil_signed_s16: Coil
):
    coil_data = CoilData(coil_signed_s16, 15.1)
    assert encoder_word_swap_false.encode(coil_data) == b"\x97\x00\x00\x00"


@pytest.mark.parametrize(
    "value",
    [
        (4),
        (30.1),
    ],
)
def test_signed_s16_encode_exception(
    value, encoder_word_swap_false: CoilDataEncoderNibeGw, coil_signed_s16: Coil
):
    with pytest.raises(EncodeException):
        coil_data = CoilData(coil_signed_s16, value)
        encoder_word_swap_false.encode(coil_data)


# Unsigned 16-bit
@pytest.fixture
def coil_unsigned_u16():
    return Coil(
        123,
        "compressor-frequency-actual-43136",
        "Compressor Frequency, Actual",
        "u16",
        factor=10,
    )


@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x01\x00\x00\x00", 0.1),
        (b"\x01\x00", 0.1),
        (b"\xff\xff", None),
    ],
)
def test_unsigned_u16_decode(
    raw_value,
    value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_unsigned_u16: Coil,
):
    assert encoder_word_swap_false.decode(coil_unsigned_u16, raw_value) == CoilData(
        coil_unsigned_u16, value
    )


@pytest.mark.parametrize(
    "value, raw_value",
    [
        (0.1, b"\x01\x00\x00\x00"),
        (25.5, b"\xff\x00\x00\x00"),
        (18.0, b"\xb4\x00\x00\x00"),
    ],
)
def test_unsigned_u16_encode(
    value,
    raw_value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_unsigned_u16: Coil,
):
    coil_data = CoilData(coil_unsigned_u16, value)
    assert encoder_word_swap_false.encode(coil_data) == raw_value


# Unsigned 16-bit word swap
@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x01\x00\x00\x00", 0.1),
        (b"\x01\x00", 0.1),
        (b"\xff\xff", None),
        (b"\xb4\x00\x00\x00", 18.0),
    ],
)
def test_unsigned_u16_word_swap_decode(
    raw_value, value, encoder_word_swap: CoilDataEncoderNibeGw, coil_unsigned_u16: Coil
):
    actual_coil_data = encoder_word_swap.decode(coil_unsigned_u16, raw_value)
    assert actual_coil_data == CoilData(coil_unsigned_u16, value)
    assert isinstance(actual_coil_data.value, float) or actual_coil_data.value is None


@pytest.mark.parametrize(
    "value, raw_value",
    [(0.1, b"\x01\x00\x00\x00"), (25.5, b"\xff\x00\x00\x00")],
)
def test_unsigned_u16_word_swap_encode(
    value, raw_value, encoder_word_swap: CoilDataEncoderNibeGw, coil_unsigned_u16: Coil
):
    coil_data = CoilData(coil_unsigned_u16, value)
    assert encoder_word_swap.encode(coil_data) == raw_value


# Signed 32-bit
@pytest.fixture
def coil_signed_s32():
    return Coil(
        43420,
        "tot-op-time-compr-eb100-ep14-43420",
        "Total compressorer operation time",
        "s32",
    )


@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x00\x002T", 21554),
        (b"\x00\x80\x00\x00", None),
    ],
)
def test_signed_s32_decode(
    raw_value,
    value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_signed_s32: Coil,
):
    actual_coil_data = encoder_word_swap_false.decode(coil_signed_s32, raw_value)
    assert actual_coil_data == CoilData(coil_signed_s32, value)
    assert isinstance(actual_coil_data.value, int) or actual_coil_data.value is None


def test_signed_s32_encode(
    encoder_word_swap_false: CoilDataEncoderNibeGw, coil_signed_s32: Coil
):
    coil_data = CoilData(coil_signed_s32, 21554)
    assert encoder_word_swap_false.encode(coil_data) == b"\x00\x002T"


# Signed 32-bit word swap
@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"(\x06\x00\x00", 1576),
        (b"\x00\x00\x00\x80", None),
    ],
)
def test_signed_s32_word_swap_decode(
    raw_value, value, encoder_word_swap: CoilDataEncoderNibeGw, coil_signed_s32: Coil
):
    assert encoder_word_swap.decode(coil_signed_s32, raw_value) == CoilData(
        coil_signed_s32, value
    )


def test_signed_s32_word_swap_encode(
    encoder_word_swap: CoilDataEncoderNibeGw, coil_signed_s32: Coil
):
    coil_data = CoilData(coil_signed_s32, 1576)
    assert encoder_word_swap.encode(coil_data) == b"(\x06\x00\x00"


# Unsigned 8-bit with mapping
@pytest.fixture
def coil_unsigned_u8_mapping():
    return Coil(
        123,
        "prio-43086",
        "Prio",
        "u8",
        factor=1,
        mappings={
            "10": "Off",
            "20": "Hot Water",
            "30": "Heat",
            "40": "Pool",
            "41": "Pool 2",
            "50": "Transfer",
            "60": "Cooling",
        },
    )


@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x0a", "OFF"),
        (b"\xff\xff", None),
    ],
)
def test_unsigned_u8_mapping_decode(
    raw_value,
    value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_unsigned_u8_mapping: Coil,
):
    assert encoder_word_swap_false.decode(
        coil_unsigned_u8_mapping, raw_value
    ) == CoilData(coil_unsigned_u8_mapping, value)


def test_unsigned_u8_mapping_decode_exception(
    encoder_word_swap_false: CoilDataEncoderNibeGw, coil_unsigned_u8_mapping: Coil
):
    with pytest.raises(DecodeException):
        encoder_word_swap_false.decode(coil_unsigned_u8_mapping, b"\x00")


@pytest.mark.parametrize(
    "value, raw_value",
    [
        ("OFF", b"\x0a\x00\x00\x00"),
        ("off", b"\x0a\x00\x00\x00"),
        ("Hot Water", b"\x14\x00\x00\x00"),
    ],
)
def test_unsigned_u8_mapping_encode(
    value,
    raw_value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_unsigned_u8_mapping: Coil,
):
    coil_data = CoilData(coil_unsigned_u8_mapping, value)
    assert encoder_word_swap_false.encode(coil_data) == raw_value


def test_unsigned_u8_mapping_encode_exception(
    encoder_word_swap_false: CoilDataEncoderNibeGw, coil_unsigned_u8_mapping: Coil
):
    coil_data = CoilData(coil_unsigned_u8_mapping, "Beer")
    with pytest.raises(EncodeException):
        encoder_word_swap_false.encode(coil_data)


# Unsigned 8-bit boolean
@pytest.fixture
def coil_unsigned_u8_boolean():
    return Coil(
        43024,
        "status-cooling-43024",
        "Status Cooling",
        "u8",
        factor=1,
        mappings={"0": "Off", "1": "On"},
    )


def test_unsigned_u8_boolean(coil_unsigned_u8_boolean: Coil):
    assert coil_unsigned_u8_boolean.is_boolean


def test_unsigned_u8_encode(
    encoder_word_swap_false: CoilDataEncoderNibeGw, coil_unsigned_u8_boolean: Coil
):
    coil_data = CoilData(coil_unsigned_u8_boolean, "On")
    assert encoder_word_swap_false.encode(coil_data) == b"\x01\x00\x00\x00"


# Unsigned 8-bit boolean with bounds
@pytest.fixture
def coil_unsigned_u8_boolean_with_bounds():
    return Coil(
        47050,
        "status-cooling-43024",
        "Periodic HW",
        "s8",
        factor=1,
        min=0,
        max=1,
    )


def test_unsigned_u8_boolean_with_bounds_is_boolean(
    coil_unsigned_u8_boolean_with_bounds: Coil,
):
    assert coil_unsigned_u8_boolean_with_bounds.is_boolean


@pytest.mark.parametrize(
    "value, raw_value",
    [
        ("On", b"\x01\x00\x00\x00"),
        ("Off", b"\x00\x00\x00\x00"),
    ],
)
def test_unsigned_u8_boolean_with_bounds_encode(
    value,
    raw_value,
    encoder_word_swap_false: CoilDataEncoderNibeGw,
    coil_unsigned_u8_boolean_with_bounds: Coil,
):
    coil_data = CoilData(coil_unsigned_u8_boolean_with_bounds, value)
    assert encoder_word_swap_false.encode(coil_data) == raw_value


@pytest.mark.parametrize("size", ["u32", "s32"])
def test_word_swap_unset(size, encoder: CoilDataEncoderNibeGw):
    coil = Coil(1, "test", "test", size)
    with pytest.raises(DecodeException):
        encoder.decode(coil, b"(\x06\x00\x00")

    with pytest.raises(EncodeException):
        encoder.encode(CoilData(coil, 1))


# Date
@pytest.fixture
def coil_date():
    return Coil(
        123,
        "holiday-48044",
        "Holiday",
        "u16",
        type="date",
    )


@pytest.mark.parametrize(
    "raw_value, value",
    [
        (b"\x01\x00", datetime.date(2007, 1, 2)),
        (b"\xbf\x17", datetime.date(2023, 8, 24)),
        (b"\xc5\x17", datetime.date(2023, 8, 30)),
    ],
)
def test_date_decode(
    raw_value, value, encoder_word_swap: CoilDataEncoderNibeGw, coil_date: Coil
):
    assert encoder_word_swap.decode(coil_date, raw_value) == CoilData(coil_date, value)


@pytest.mark.parametrize(
    "value, raw_value",
    [
        (datetime.date(2007, 1, 2), b"\x01\x00\x00\x00"),
        (datetime.date(2023, 8, 24), b"\xbf\x17\x00\x00"),
        (datetime.date(2023, 8, 30), b"\xc5\x17\x00\x00"),
    ],
)
def test_date_encode(
    value, raw_value, encoder_word_swap: CoilDataEncoderNibeGw, coil_date: Coil
):
    coil_data = CoilData(coil_date, value)
    assert encoder_word_swap.encode(coil_data) == raw_value



from construct import Transformed


def swapwords(data):
    r"""
    Performs a 2 byte word swap on byte-string.

    Example:

        >>> swapwords(b'abcd')
        b'cdab'
    """
    if len(data) % 2:
        raise ValueError(f"data length {len(data)} must be a multiple of 2")

    return b"".join(data[p : p + 2] for p in range(len(data) - 2, -1, -2))


def WordSwapped(subcon):
    r"""
    Swaps the 2-byte word order within boundaries of given subcon. Requires a fixed sized subcon.

    :param subcon: Construct instance

    :raises SizeofError: ctor or compiler could not compute subcon size
    :raises ValueError: subcon size is not multiple of 2

    See :class:`~construct.core.Transformed` and :class:`~construct.core.Restreamed` for raisable exceptions.

    Example::

        Int24ul <--> ByteSwapped(Int24ub) <--> BytesInteger(3, swapped=True) <--> ByteSwapped(BytesInteger(3))
    """

    size = subcon.sizeof()
    return Transformed(subcon, swapwords, size, swapwords, size)



from __future__ import annotations

from ast import literal_eval
import asyncio
from contextlib import AbstractAsyncContextManager
import io
import logging
import re
from typing import IO

import asyncclick as click
from construct import (
    Const,
    ConstructError,
    GreedyRange,
    Int8ul,
    RawCopy,
    Select,
    Struct,
    Terminated,
)

from ..coil import CoilData
from ..connection import Connection
from ..connection.modbus import Modbus
from ..connection.nibegw import NibeGW, Request, Response
from ..exceptions import NibeException
from ..heatpump import HeatPump, Model

Ack = Struct("fields" / RawCopy(Struct("Ack" / Const(0x06, Int8ul))))

Nak = Struct("fields" / RawCopy(Struct("Nak" / Const(0x15, Int8ul))))

Block = Select(
    Terminated,
    Response,
    Request,
    Ack,
    Nak,
)

Stream = GreedyRange(Block)


@click.group()
@click.option("-v", "--verbose", count=True)
async def cli(verbose: int):
    if verbose == 0:
        log_level = logging.WARNING
    elif verbose == 1:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG

    logging.basicConfig(
        format="[%(levelname)-8s] %(message)s",
        level=log_level,
    )
    logging.log(logging.INFO, "Log level set to %r", log_level)


class ConnectionContext(AbstractAsyncContextManager):
    def __init__(self, heatpump: HeatPump, connection: Connection) -> None:
        self.heatpump = heatpump
        self.connection = connection

    async def __aenter__(self) -> ConnectionContext:
        await self.connection.start()
        return self

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        await self.connection.stop()


@cli.group(chain=True, help="Connect using nibegw protocol")
@click.argument("remote_ip", type=str)
@click.option("--listening_ip", type=str)
@click.option("--listening_port", type=int, default=10090)
@click.option("--remote_read_port", type=int, default=10091)
@click.option("--remote_write_port", type=int, default=10092)
@click.option(
    "--model",
    type=click.Choice([model.name for model in Model.__members__.values()]),
    default=None,
)
@click.pass_context
async def nibegw(
    ctx: click.Context,
    remote_ip: str,
    listening_port: int,
    listening_ip: str | None,
    remote_read_port: int,
    remote_write_port: int,
    model: str | None,
):
    heatpump = HeatPump()
    if model:
        heatpump.model = Model[model]
        await heatpump.initialize()

    connection = NibeGW(
        heatpump=heatpump,
        remote_ip=remote_ip,
        listening_port=listening_port,
        listening_ip=listening_ip,
        remote_read_port=remote_read_port,
        remote_write_port=remote_write_port,
    )

    ctx.obj = await ctx.with_async_resource(ConnectionContext(heatpump, connection))


@cli.group(chain=True, help="Connect using modbus protocol")
@click.argument("remote_ip", type=str)
@click.option("--remote_port", type=int, default=502)
@click.option("--slave_id", type=int, default=1)
@click.option(
    "--model",
    type=click.Choice([model.name for model in Model]),
    default=Model.F1155.name,
)
@click.pass_context
async def modbus(
    ctx: click.Context, remote_ip: str, remote_port: str, model: str, slave_id: int
):
    heatpump = HeatPump(Model[model])
    await heatpump.initialize()
    connection = Modbus(
        heatpump=heatpump,
        url=f"tcp://{remote_ip}:{remote_port}",
        slave_id=slave_id,
    )

    ctx.obj = await ctx.with_async_resource(ConnectionContext(heatpump, connection))


def add_connect_command(command: click.Command):
    nibegw.add_command(command)
    modbus.add_command(command)


@click.command(help="Monitor data sent by pump out of band")
@click.pass_obj
async def monitor(obj: ConnectionContext):
    def on_coil_update(coil_data: CoilData):
        click.echo(coil_data)

    obj.heatpump.subscribe(HeatPump.COIL_UPDATE_EVENT, on_coil_update)

    while True:
        await asyncio.sleep(1)


add_connect_command(monitor)


@click.command()
@click.pass_obj
async def product(obj: ConnectionContext):
    product_info = await obj.connection.read_product_info()
    click.echo(product_info)


add_connect_command(product)


@click.command()
@click.pass_obj
@click.argument("parameter", type=int)
async def read(obj: ConnectionContext, parameter: int, **kwargs):
    coil = obj.heatpump.get_coil_by_address(parameter)
    click.echo(await obj.connection.read_coil(coil))


add_connect_command(read)


@click.command()
@click.pass_obj
@click.argument("parameter", type=int)
@click.argument("value", type=str)
async def write(obj: ConnectionContext, parameter: int, value: str, **kwargs):
    coil = obj.heatpump.get_coil_by_address(parameter)
    if not coil.mappings:
        value = float(value)

    coil_data = CoilData(coil, value)
    await obj.connection.write_coil(coil_data)
    click.echo(coil_data)


add_connect_command(write)


@cli.command()
@click.argument("data", type=str)
@click.option("--type", type=click.Choice(["hex", "bytes"]), default="hex")
async def parse_data(data: str, type: str):
    if type == "hex":
        raw = bytes.fromhex(data)
    elif type == "bytes":
        raw = bytes(literal_eval(data))

    try:
        request = Block.parse(raw)
        click.echo(request)
    except ConstructError as exception:
        click.echo(f"Failed to parse: {exception}", err=True)


def read_bytes_socat(file: IO):
    lines: list[str] = file.readlines()

    for line in lines:
        if line.startswith("> "):
            continue
        yield from bytes.fromhex(line)


RE_NIBEPI_LOG = re.compile(r"Serial: \[((?:[0-9]+,?)+)\]")


def read_bytes_nibepi(file: IO):
    lines: list[str] = file.readlines()
    for line in lines:
        data = RE_NIBEPI_LOG.match(line)
        if data:
            yield from map(int, data.group(1).split(","))


def parse_stream(stream: io.RawIOBase):
    while block := Block.parse_stream(stream):
        yield block


@cli.command()
@click.argument("file", type=click.File())
@click.option("--type", type=click.Choice(["hex", "nibepi"]), default="hex")
def parse_file(file: IO, type: str):
    if type == "hex":
        data = read_bytes_socat(file)
    elif type == "nibepi":
        data = read_bytes_nibepi(file)
    else:
        raise ValueError("invalid argument")

    with io.BytesIO(bytes(data)) as stream:
        for packet in parse_stream(stream):
            click.echo(packet.fields.value)

        remaining = stream.read()
        if remaining:
            click.echo(f"Remaining: {stream.read()}")


def main():
    try:
        cli()
    except NibeException as exception:
        click.echo(repr(exception))
    except (KeyboardInterrupt, SystemExit):
        pass


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass

from .heatpump import Series


@dataclass
class CoilGroup:
    name: str


@dataclass
class UnitCoilGroup(CoilGroup):
    prio: int
    cooling_with_room_sensor: int
    alarm: int
    alarm_reset: int


@dataclass
class CompressorCoilGroup(CoilGroup):
    pass


@dataclass
class ClimateCoilGroup(CoilGroup):
    active_accessory: int | None
    current: int
    setpoint_heat: int
    setpoint_cool: int
    mixing_valve_state: int
    use_room_sensor: int


@dataclass
class WaterHeaterCoilGroup(CoilGroup):
    hot_water_load: int
    hot_water_comfort_mode: int
    start_temperature: dict[str, int]
    stop_temperature: dict[str, int]
    active_accessory: int | None
    temporary_lux: int | None


@dataclass
class FanCoilGroup(CoilGroup):
    speed: int | None
    speeds: dict[str, int]


_UNIT_COILGROUPS_F = {
    "main": UnitCoilGroup(
        name="Main",
        prio=43086,
        cooling_with_room_sensor=47340,
        alarm=45001,
        alarm_reset=45171,
    ),
}

_UNIT_COILGROUPS_S = {
    "main": UnitCoilGroup(
        name="Main",
        prio=31029,
        cooling_with_room_sensor=40171,
        alarm=31976,
        alarm_reset=40023,
    ),
}

UNIT_COILGROUPS = {
    Series.F: _UNIT_COILGROUPS_F,
    Series.S: _UNIT_COILGROUPS_S,
}

_CLIMATE_COILGROUPS_F = {
    "s1": ClimateCoilGroup(
        name="Climate System S1",
        current=40033,
        setpoint_heat=47398,
        setpoint_cool=48785,
        mixing_valve_state=43096,
        active_accessory=None,
        use_room_sensor=47394,
    ),
    "s2": ClimateCoilGroup(
        name="Climate System S2",
        current=40032,
        setpoint_heat=47397,
        setpoint_cool=48784,
        mixing_valve_state=43095,
        active_accessory=47302,
        use_room_sensor=47393,
    ),
    "s3": ClimateCoilGroup(
        name="Climate System S3",
        current=40031,
        setpoint_heat=47396,
        setpoint_cool=48783,
        mixing_valve_state=43094,
        active_accessory=47303,
        use_room_sensor=47392,
    ),
    "s4": ClimateCoilGroup(
        name="Climate System S4",
        current=40030,
        setpoint_heat=47395,
        setpoint_cool=48782,
        mixing_valve_state=43093,
        active_accessory=47304,
        use_room_sensor=47391,
    ),
}

_CLIMATE_COILGROUPS_S = {
    "s1": ClimateCoilGroup(
        name="Climate System S1",
        current=30027,
        setpoint_heat=40207,
        setpoint_cool=40989,
        mixing_valve_state=31034,
        active_accessory=None,
        use_room_sensor=40203,
    ),
    "s2": ClimateCoilGroup(
        name="Climate System S2",
        current=30026,
        setpoint_heat=40206,
        setpoint_cool=40988,
        mixing_valve_state=31033,
        active_accessory=None,
        use_room_sensor=40202,
    ),
    "s3": ClimateCoilGroup(
        name="Climate System S3",
        current=30025,
        setpoint_heat=40205,
        setpoint_cool=40987,
        mixing_valve_state=31032,
        active_accessory=None,
        use_room_sensor=40201,
    ),
    "s4": ClimateCoilGroup(
        name="Climate System S4",
        current=30024,
        setpoint_heat=40204,
        setpoint_cool=40986,
        mixing_valve_state=31031,
        active_accessory=None,
        use_room_sensor=40200,
    ),
}

CLIMATE_COILGROUPS = {
    Series.F: _CLIMATE_COILGROUPS_F,
    Series.S: _CLIMATE_COILGROUPS_S,
}

_WATER_HEATER_COILGROUPS_F = {
    "hw1": WaterHeaterCoilGroup(
        name="Hot Water",
        hot_water_load=40014,
        hot_water_comfort_mode=47041,
        start_temperature={
            "ECONOMY": 47045,
            "NORMAL": 47044,
            "LUXURY": 47043,
        },
        stop_temperature={
            "ECONOMY": 47049,
            "NORMAL": 47048,
            "LUXURY": 47047,
        },
        active_accessory=None,
        temporary_lux=48132,
    ),
}

_WATER_HEATER_COILGROUPS_S = {
    "hw1": WaterHeaterCoilGroup(
        name="Hot Water",
        hot_water_load=30010,
        hot_water_comfort_mode=31039,
        start_temperature={
            "LOW": 40061,
            "NORMAL": 40060,
            "HIGH": 40059,
        },
        stop_temperature={
            "LOW": 40065,
            "NORMAL": 40064,
            "HIGH": 40063,
        },
        active_accessory=None,
        temporary_lux=None,
    ),
}

WATER_HEATER_COILGROUPS = {
    Series.F: _WATER_HEATER_COILGROUPS_F,
    Series.S: _WATER_HEATER_COILGROUPS_S,
}

_FAN_COILGROUPS_F = {
    "exhaust": FanCoilGroup(
        name="Exhaust",
        speed=47260,
        speeds={
            "0": 47265,
            "1": 47264,
            "2": 47263,
            "3": 47262,
            "4": 47261,
        },
    ),
    "supply": FanCoilGroup(
        name="Supply",
        speed=47260,
        speeds={
            "0": 47270,
            "1": 47269,
            "2": 47268,
            "3": 47267,
            "4": 47266,
        },
    ),
}

_FAN_COILGROUPS_S = {}

FAN_COILGROUPS = {
    Series.F: _FAN_COILGROUPS_F,
    Series.S: _FAN_COILGROUPS_S,
}



from exceptiongroup import ExceptionGroup


class NibeException(Exception):
    pass


class AddressInUseException(NibeException):
    pass


class CoilNotFoundException(NibeException):
    pass


class DecodeException(NibeException):
    pass


class ValidationError(NibeException):
    pass


class NoMappingException(ValidationError):
    pass


class EncodeException(NibeException):
    pass


class WriteException(NibeException):
    pass


class WriteIOException(WriteException):
    """Use this and child exceptions if IO has failed and you want to retry."""

    pass


class CoilWriteSendException(WriteIOException):
    pass


class WriteTimeoutException(WriteIOException):
    pass


class ReadException(NibeException):
    pass


class ReadIOException(ReadException):
    """Use this and child exception if IO has failed and you want to retry."""

    pass


class ReadExceptionGroup(ExceptionGroup, ReadException):
    def __str__(self) -> str:
        messages = ", ".join(str(exception) for exception in self.exceptions)
        return f"{self.message} ({messages})"


class ReadSendException(ReadIOException):
    pass


class ReadTimeoutException(ReadIOException):
    pass


class ProductInfoReadTimeoutException(ReadIOException):
    pass


class ModelIdentificationFailed(NibeException):
    pass


class ModbusUrlException(NibeException, ValueError):
    pass



import asyncio
from dataclasses import dataclass
from enum import Enum, auto
from importlib.resources import files
import json
import logging
from os import PathLike
from typing import Dict, Optional, Union

from nibe.coil import Coil, CoilData
from nibe.event_server import EventServer
from nibe.exceptions import CoilNotFoundException, ModelIdentificationFailed

logger = logging.getLogger("nibe").getChild(__name__)


class Series(Enum):
    """Series enum class"""

    CUSTOM = auto()
    F = auto()
    S = auto()


class Model(Enum):
    """Model enum class"""

    F1155 = "f1155_f1255", Series.F
    F1255 = "f1155_f1255", Series.F

    S1155 = "s1155_s1255", Series.S
    S1255 = "s1155_s1255", Series.S

    S1256 = "s1155_s1255", Series.S

    F1145 = "f1145_f1245", Series.F
    F1245 = "f1145_f1245", Series.F

    F1345 = "f1345", Series.F
    F1355 = "f1355", Series.F

    F730 = "f730", Series.F
    F750 = "f750", Series.F

    F370 = "f370_f470", Series.F
    F470 = "f370_f470", Series.F

    S320 = "s320_s325", Series.S
    S325 = "s320_s325", Series.S

    S735 = "s735", Series.S

    S2125 = "s2125", Series.S

    SMO20 = "smo20", Series.F
    SMO40 = "smo40", Series.F

    SMOS40 = "smos40", Series.S

    VVM225 = "vvm225_vvm320_vvm325", Series.F
    VVM320 = "vvm225_vvm320_vvm325", Series.F
    VVM325 = "vvm225_vvm320_vvm325", Series.F

    VVM310 = "vvm310_vvm500", Series.F
    VVM500 = "vvm310_vvm500", Series.F

    CUSTOM = "custom", Series.CUSTOM

    data_file: Union[str, bytes, PathLike[str], PathLike[bytes]]
    series: Series

    def __new__(cls, data_file: str, series: Series):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        obj.data_file = data_file
        obj.series = series
        return obj

    def get_coil_data(self):
        """Get coil data for model"""
        if self == Model.CUSTOM:
            with open(self.data_file) as fh:
                return json.load(fh)
        else:
            with files("nibe.data").joinpath(f"{self.data_file}.json").open("r") as fh:
                return json.load(fh)

    @classmethod
    def keys(cls):
        """Get all keys of the enum class"""
        return cls.__members__.keys()


@dataclass
class ProductInfo:
    """Product info class"""

    model: str
    firmware_version: int

    def identify_model(self) -> Model:
        """Identify model from product info

        :raises ModelIdentificationFailed: When model cannot be identified"""
        for key in Model.keys():
            if key in self.model.upper():
                return getattr(Model, key)

        raise ModelIdentificationFailed(f'Unable to identify model from "{self.model}"')


class HeatPump(EventServer):
    """Heat pump class"""

    COIL_UPDATE_EVENT = "coil_update"

    _address_to_coil: Dict[str, Coil]
    _name_to_coil: Dict[str, Coil]
    word_swap: Optional[bool] = None
    _product_info: Union[ProductInfo, None] = None
    _model: Optional[Model] = None

    def __init__(self, model: Optional[Model] = None):
        super().__init__()

        self._address_to_coil = {}
        self._name_to_coil = {}

        if model is not None:
            self.model = model

    @property
    def model(self) -> Union[Model, None]:
        """Returns the model of the heat pump"""
        return self._model

    @model.setter
    def model(self, model: Model):
        """Sets the model of the heat pump"""
        assert isinstance(model, Model), "Passed argument is not of a Model type"

        self._model = model

    @property
    def series(self) -> Series:
        """Returns the series of the heat pump"""
        assert self._model
        return self._model.series

    @property
    def product_info(self) -> Union[ProductInfo, None]:
        """Returns the product info of the heat pump"""
        return self._product_info

    @product_info.setter
    def product_info(self, product_info: ProductInfo):
        """Sets the product info of the heat pump"""
        assert isinstance(
            product_info, ProductInfo
        ), "Passed argument is not of a ProductInfo type"

        self._product_info = product_info

    async def _load_coils(self):
        assert isinstance(self._model, Model), "Model is not set"
        data = await asyncio.get_running_loop().run_in_executor(
            None, self._model.get_coil_data
        )

        self._address_to_coil = {}
        for k, v in data.items():
            try:
                self._address_to_coil[k] = Coil(address=int(k), **v)
            except (AssertionError, TypeError) as e:
                logger.warning(f"Failed to register coil {k}: {e}")
        self._name_to_coil = {c.name: c for _, c in self._address_to_coil.items()}

    async def initialize(self):
        """Initialize the heat pump"""
        if not isinstance(self._model, Model) and isinstance(
            self._product_info, ProductInfo
        ):
            self.model = self._product_info.identify_model()

        assert isinstance(
            self._model, Model
        ), "Model is not set and product info is not available"

        await self._load_coils()

    def get_coils(self) -> list[Coil]:
        """Returns a list of all coils"""
        return list(self._address_to_coil.values())

    def get_coil_by_address(self, address: Union[int, str]) -> Coil:
        """Returns a coil by address

        :raises CoilNotFoundException: if coil is not found
        """
        try:
            return self._address_to_coil[str(address)]
        except KeyError:
            raise CoilNotFoundException(f"Coil with address {address} not found")

    def get_coil_by_name(self, name: str) -> Coil:
        """Returns a coil by name

        :raises CoilNotFoundException: if coil is not found
        """
        try:
            return self._name_to_coil[str(name)]
        except KeyError:
            raise CoilNotFoundException(f"Coil with name '{name}' not found")

    def notify_coil_update(self, coil_data: CoilData):
        """Notifies listeners about coil update"""
        self.notify_event_listeners(self.COIL_UPDATE_EVENT, coil_data)



import unittest
from unittest.mock import Mock

import pytest

from nibe.coil import CoilData
from nibe.exceptions import CoilNotFoundException, ModelIdentificationFailed
from nibe.heatpump import HeatPump, Model, ProductInfo, Series


class HeatpumpTestCase(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.heat_pump = HeatPump(Model.F1255)
        await self.heat_pump.initialize()

        assert len(self.heat_pump._address_to_coil) > 100

    def test_get_coils(self):
        coils = self.heat_pump.get_coils()
        assert isinstance(coils, list)

    def test_get_coil_by_returns_same(self):
        coil_address = 40004

        a = self.heat_pump.get_coil_by_address(coil_address)
        b = self.heat_pump.get_coil_by_address(str(coil_address))

        assert coil_address == a.address

        assert a is b

        c = self.heat_pump.get_coil_by_name("bt1-outdoor-temperature-40004")

        assert a is c

    def test_get_missing_coil_raises_exception(self):
        with pytest.raises(CoilNotFoundException):
            self.heat_pump.get_coil_by_address(0xFFFF)

        with pytest.raises(CoilNotFoundException):
            self.heat_pump.get_coil_by_name("no-beer-today")

    def test_listener(self):
        mock = Mock()
        coil = self.heat_pump.get_coil_by_address(40004)
        self.heat_pump.subscribe(self.heat_pump.COIL_UPDATE_EVENT, mock)

        mock.assert_not_called()

        coil_data = CoilData(coil, 14)
        self.heat_pump.notify_coil_update(coil_data)

        mock.assert_called_with(coil_data)

    def test_listener_with_exception(self):
        mock = Mock(side_effect=Exception("Test exception that needs to be logged"))
        coil = self.heat_pump.get_coil_by_address(40004)
        self.heat_pump.subscribe(self.heat_pump.COIL_UPDATE_EVENT, mock)

        coil_data = CoilData(coil, 14)
        self.heat_pump.notify_coil_update(
            coil_data
        )  # Error should be logged but not thrown out


class HeatpumpIntialization(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.heat_pump = HeatPump()

    async def test_initalize_with_model(self):
        self.heat_pump.model = Model.F1255
        await self.heat_pump.initialize()
        self.heat_pump.get_coil_by_address(43420)
        assert self.heat_pump.model is Model.F1255

    async def test_initalize_with_product_info(self):
        product_info = ProductInfo("F1255-12 R", 0)
        self.heat_pump.product_info = product_info
        await self.heat_pump.initialize()
        self.heat_pump.get_coil_by_address(43420)
        assert self.heat_pump.product_info is product_info

    async def test_initalization_failed(self):
        with pytest.raises(AssertionError):
            await self.heat_pump.initialize()


@pytest.mark.parametrize(
    "model,series",
    [
        (Model.F370, Series.F),
        (Model.F730, Series.F),
        (Model.F1145, Series.F),
        (Model.F1245, Series.F),
        (Model.F1155, Series.F),
        (Model.F1255, Series.F),
        (Model.F1355, Series.F),
        (Model.SMO20, Series.F),
        (Model.SMO40, Series.F),
        (Model.VVM225, Series.F),
        (Model.VVM320, Series.F),
        (Model.VVM325, Series.F),
        (Model.VVM310, Series.F),
        (Model.VVM500, Series.F),
        (Model.SMOS40, Series.S),
        (Model.S320, Series.S),
        (Model.S325, Series.S),
        (Model.S735, Series.S),
        (Model.S1155, Series.S),
        (Model.S1255, Series.S),
        (Model.S1256, Series.S),
    ],
)
def test_series(model: Model, series: Series):
    heat_pump = HeatPump(model)
    assert heat_pump.series == series


class ProductInfoTestCase(unittest.TestCase):
    def test_identify_model(self):
        product_info = ProductInfo("F1255-12 R", 0)

        assert product_info.identify_model() == Model.F1255

        product_info = ProductInfo("F1155-16", 0)

        assert product_info.identify_model() == Model.F1155

    def test_identify_model_error(self):
        product_info = ProductInfo("Tehowatti Air", 0)

        with pytest.raises(ModelIdentificationFailed):
            product_info.identify_model()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Iterable

from nibe.coil import Coil, CoilData
from nibe.exceptions import ReadException, ReadExceptionGroup
from nibe.heatpump import HeatPump, ProductInfo, Series

DEFAULT_TIMEOUT: float = 5
READ_PRODUCT_INFO_TIMEOUT: float = 20  # Product info message is sent every 15 seconds


class Connection(ABC):
    """Base class for all connection methods."""

    async def start(self):  # noqa: B027
        """Start the connection."""
        pass

    async def stop(self):  # noqa: B027
        """Close the connection."""
        pass

    @abstractmethod
    async def read_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> CoilData:
        """Read a coil data from the heatpump.

        :raises ReadIOException: If failed to read coil data due to IO error (will retry).
        :raises ReadException: If failed to read coil data due to other error (will not retry).
        """
        pass

    async def read_coils(
        self, coils: Iterable[Coil], timeout: float = DEFAULT_TIMEOUT
    ) -> AsyncIterator[CoilData]:
        """Read multiple coil data from the heatpump.

        :raises ReadExceptionGroup: If one or more coils failed to read."""
        exceptions = []
        for coil in coils:
            try:
                yield await self.read_coil(coil, timeout)
            except ReadException as exception:
                exceptions.append(exception)
        if exceptions:
            raise ReadExceptionGroup("Failed to read some or all coils", exceptions)

    @abstractmethod
    async def write_coil(
        self, coil_data: CoilData, timeout: float = DEFAULT_TIMEOUT
    ) -> None:
        """Write a coil data to the heatpump.

        :raises WriteIOException: If failed to write coil data due to IO error (will retry).
        :raises WriteException: If failed to write coil data due to other error (will not retry).
        """
        pass

    async def read_product_info(
        self, timeout: float = READ_PRODUCT_INFO_TIMEOUT
    ) -> ProductInfo:
        """Read product info from the heatpump.

        :raises ReadIOException: If failed to read product info in time."""
        raise NotImplementedError(
            "read_product_info method is not implemented for this connection method"
        )

    @abstractmethod
    async def verify_connectivity(self):
        """Verify that we have functioning communication.

        :raises NibeException: If failed to verify connectivity."""
        pass


async def verify_connectivity_read_write_alarm(
    connection: Connection, heatpump: HeatPump
):
    """Verify that we have functioning communication.

    To verify connection, we read the alarm reset field and write it as 0
    this will be ignored by the pump, this will throw exceptions on failure.
    """
    if heatpump.series == Series.S:
        coil = heatpump.get_coil_by_name("reset-alarm-40023")
    else:
        coil = heatpump.get_coil_by_name("alarm-reset-45171")

    coil_data = await connection.read_coil(coil)
    value: str | int = 0
    if coil.mappings:
        value = coil.mappings[str(value)]
    coil_data.value = value
    await connection.write_coil(coil_data)
